/* bdb-chol.pn.c */

/* This program is the host processor software for a parallel version
of the Shur complement Choleski factorization algorithm for
block-diagonal-bordered form sparse symmetric matrices.  */

/* special considerations for this implementation: 

1) Matrices are in symmetric block-diagonal-bordered form

2) any number of processors can be specified up to an including the
total number of processors on the CM-5 

3) it is assumed that ordering and load balancing are performed a
priori

4) the last block is considered dense, and a dense blocked symmetric
Choleski algorithm is used to factor this portion of the matrix

*/

/* Naming convention used throughout this software

i - row number
j - column number
k - number along the diagonal

Lij   - any value in a strictly lower diagonal matrix block
Lkk   - value of the kth element on the lower diagonal
        stored in the location diag[k].Value
Lix   - row in a strictly lower diagonal matrix block 
        (ith row contributing to update of Lij)
Ljx   - row in a strictly lower diagonal matrix block 
        (jth row contributing to update of Lij)
        Because matrix is symmetric, this row would 
        correspond to the column Uxj in L^T
Lxk   - column in a strictly lower diagonal matrix block
        only accessed when updating a column

LBij  - any value in a lower  matrix border
LBix  - row in a lower matrix border - (ith row contributing 
        to update of LBij)
LBjx  - row in a matrix  border (jth row contributing 
        to update of LBij)   Because matrix is symmetric, this row would 
        correspond to the border column UBxj in L^T
LBxk  - column in a lower matrix border
        only accessed when updating a column

locLijLkj    - function to find the implicit storage locations 
               of Lij and Lkj, where i and k are given, and 
               (Lij != 0.) && (Lkj != 0.) && (j < k)
locLBijLkj   - function to find the implicit storage locations 
               of LBij and Ujk, where i and k are given, and 
               (LBij != 0.) && (Ujk != 0.) && (j < k)
LBixLBkx    -  function to find the sum of the products of values
               in partial row LBix and partial row LBkx
locLBixLBkx - function to find the implicit storage locations 
               of LBij and LBkj, where i and k are given, and 
               (LBij != 0.) && (LBkj != 0.)
*/

#include "AIXtranslate.h"
#include <stdio.h>
#include <sys/types.h>
#include <math.h>
#include <malloc.h>
/* #include <cm/cmmd.h> */
#include "bdb_struct_par_chol.h"
#include "bdb_chol_MSG_NUM.h"
/*---------------------------------------------------------*/
#define YES        1
#define NO         0
#define EPSILON    0.0000001
/*---------------------------------------------------------*/
typedef struct Node_Timer{int pn_id;
                          double elapsed;
                          double busy;
                          double idle;} node_timer;
node_timer node_tim;
/*---------------------------------------------------------*/
Block       *blocks;     /* blocks[]    */
Diag        *diag;       /* diag[]      */
BlockRow    *blockrow;   /* blockrow[]  */
LowerBorder *lborder;    /* lborder[]   */
BorderRow   *borderrow;  /* borderrow[] */

LBixLBkxBuffer **LBixLBkxBuf;  /* LBixLBkxBuf[][] */
FRBBuffer      **FrbBuf;       /* FrbBuf[][] */

double         *x;         /* x[]         */
double         *y;         /* y[]         */
double         *b;         /* b[]         */

int            n;          /* matrix dimension       */
int            nblocks;    /* # of blocks            */
int            nblockrow;  /* dimension of blockrow  */
int            nlborder;   /* dimension of lborder   */
int            nborderrow; /* dimension of borderrow */

int            nLBixLBkxBuf;  /* dimension of LBixLBkxBuf  */

int            nlastblock;
int            lastblockstart;
int            lastblockend;

int            nprocs;
int            procnum;
int            bytesreceived;
int            bytessent;
/*---------------------------------------------------------*/
typedef struct BBuf{int nprocs;
                    int n;
                    int nblocks;
                    int nblockrow;
                    int nlborder; 
                    int nborderrow;
                    int ln;
                    int nb;} BBuf;
BBuf bbuf;
/*---------------------------------------------------------*/

double *la;   /* la[] stored in column-major order (ie., a FORTRAN matrix) */
double *lx;   /* lx[] */
double *ly;   /* ly[] */
double *lb;   /* lb[] */
int    ln;    /* dimension of the last block */
int    nb;    /* block-width */

/* variables for kij-saxpy code */
int    lda,info;
char   *buf2; 
int    maxbufsize;     /* max size of communication buffer    */

int    nbks;
int    bksthisproc;
int    maxbksperproc;
int    colsthisproc;

/* function prototypes */
void InputData();
void Node_Exit();
void Factor();
  void FactorIndependentBlocks();
    double UpdateLkk();
    void UpdateLxk();
      void locLixLjx();
    void UpdateLBxk();
      void locLBixLjx();
  void UpdateLastBlock();
      void  LBixLBkx();
      void locLBixLBkx();
  void FactorLastBlock();
    extern void hlltbk();
      void bcastdoublerc();
      void bcastdoublesd();
      int nextinring();
      int previnring();
void ForwardReduction();
  void ForwardReduceBlocks();
  void ForwardReduceBorder();
  void ForwardReduceLastBlock();
    extern void frmbk();
void BackwardSubstitution();
  void BackSubLastBlock();
    extern void bsmbk();
  void BackSubBlocks();

main(argc,argv)
  int argc;
  char **argv;
{
  int i;
  struct node_timer{int pn_id;
                    double elapsed;
                    double busy;
                    double idle;} node_tim;
#include "AIX.h"

  CMMD_enable_host();
  enable_diag_write();
  procnum = CMMD_self_address();

printf("InputData() \n");
fflush(stdout);
  InputData();

printf("Factor \n");
fflush(stdout);
  Factor();

printf("ForwardReduction \n");
fflush(stdout);
     ForwardReduction();

printf("BackwardSubstitution \n");
fflush(stdout);
  BackwardSubstitution();

} /* end main() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void InputData()
{
  int i,j,k,l,m,bufsize,bytesrec;
  double e1,e2;
  LBixLBkxBuffer *LBixLBkxBuftmp;
  FRBBuffer *FrbBuftmp;
#include "AIX.h"

/* (I) set up variables at this processor */ 
  CMMD_receive_bc_from_host(&bbuf,(8*sizeof(int)));
  ln     = bbuf.ln;
  nb     = bbuf.nb;
  nprocs = bbuf.nprocs;
printf("ln=%d - nb=%d - nprocs=%d - procnum=%d  \n",ln,nb,nprocs,procnum);

if (procnum < nprocs)
 {
  nbks = ((ln+nb-1)/nb);
  bksthisproc = (nbks/nprocs);
    if (procnum < (nbks%nprocs)) bksthisproc++;
  maxbksperproc = (nbks+nprocs-1)/nprocs;
  colsthisproc = (bksthisproc*nb);
  if ( (procnum == (((nbks-1)%nprocs))) && (ln < (nbks*nb)) ) 
    colsthisproc -= ((nbks*nb)-ln);

printf("nbks=%d - bksthisproc=%d - maxbksperproc=%d - colsthisproc=%d   \n",
            nbks,bksthisproc,maxbksperproc,colsthisproc);
fflush(stdout);

/* (1) receive the block identifer array of structures */
  CMMD_receive(CMMD_host_node(),RECEIVE_NPBLOCKS,&nblocks,sizeof(int));

printf("receive the block identifer array of structures --- nblocks=%d \n",nblocks);
fflush(stdout);

  blocks = (Block *) calloc((nblocks+1),sizeof(Block));
  if (blocks == NULL)
   {
    printf("can't allocate memory for blocks[] \n");
    Node_Exit(1);
   } /* end if */
  CMMD_receive(CMMD_host_node(),RECEIVE_PBLOCKS,blocks,((nblocks+1)*sizeof(Block)));
/*
if (procnum == 1) {
  for (i=1;i<=nblocks;i++) {
    printf("%d \t%d %d %d %d %d \n",i,blocks[i].BlockID,blocks[i].BlockSz,
               blocks[i].ProcNum,blocks[i].DiagPtr,blocks[i].LBPtr);
    fflush(stdout); }}
*/
/* (2) receive the matrix diagonal array of structures */
  CMMD_receive(CMMD_host_node(),RECEIVE_NP,&n,sizeof(int));

printf("receive the matrix diagonal array of structures --- n=%d \n",n);
fflush(stdout);

  diag = (Diag *) calloc((n+1),sizeof(Diag));
  if (diag == NULL)
   {
    printf("can't allocate memory for diag[] \n");
    Node_Exit(1);
   } /* end if */
  CMMD_receive(CMMD_host_node(),RECEIVE_PDIAG,diag,((n+1)*sizeof(Diag)));
/*
if (procnum == 1) {
  for (i=1;i<=n;i++) {
    printf("%d \t%lf %d %d %d \n",i,diag[i].Value,diag[i].PtrLix,
             diag[i].NextLxk,diag[i].NextLBxk);
    fflush(stdout); }}
*/
/* (3) receive the block row storage array of structures */
  CMMD_receive(CMMD_host_node(),RECEIVE_NPBLOCKROW,&nblockrow,sizeof(int));

printf("receive the block row storage array of structures --- nblockrow=%d \n",nblockrow);
fflush(stdout);

  blockrow = (BlockRow *) calloc((nblockrow+1),sizeof(BlockRow));
  if (blockrow == NULL)
   {
    printf("can't allocate memory for blockrow[] \n");
    Node_Exit(1);
   } /* end if */
  CMMD_receive(CMMD_host_node(),RECEIVE_PBLOCKROW,blockrow,((nblockrow+1)*sizeof(BlockRow)));
/*
if (procnum == 1) {
  for (i=1;i<=nblockrow;i++) {
    printf("%d \t%lf %d %d %d \n",i,blockrow[i].Value,blockrow[i].i,
                    blockrow[i].j,blockrow[i].NextLxk);
    fflush(stdout); }}
*/
/* (5) receive the lower border row storage array of structures */
  CMMD_receive(CMMD_host_node(),RECEIVE_NPLBORDER,&nlborder,sizeof(int));

printf("receive the lower border row storage array of structures --- nlborder=%d \n",nlborder);
fflush(stdout);

  lborder = (LowerBorder *) calloc((nlborder+1),sizeof(LowerBorder));
  if (lborder == NULL)
   {
    printf("can't allocate memory for lborder[] \n");
    Node_Exit(1);
   } /* end if */
  CMMD_receive(CMMD_host_node(),RECEIVE_PLBORDER,lborder,((nlborder+1)*sizeof(LowerBorder)));
/*
if (procnum == 1) {
  for (i=1;i<=nlborder;i++) {
    printf("%d \t%d %d \n",i,lborder[i].i,lborder[i].PtrLBix);
    fflush(stdout); }}
*/
/* (7) receive the border row storage array of structures */
  CMMD_receive(CMMD_host_node(),RECEIVE_NPBORDERROW,&nborderrow,sizeof(int));

printf("receive the border row storage array of structures --- nborderrow=%d \n",nborderrow);
fflush(stdout);

  borderrow = (BorderRow *) calloc((nborderrow+1),sizeof(BorderRow));
  if (borderrow == NULL)
   {
    printf("can't allocate memory for borderrow[] \n");
    Node_Exit(1);
   } /* end if */
  CMMD_receive(CMMD_host_node(),RECEIVE_PBORDERROW,borderrow,((nborderrow+1)*sizeof(BorderRow)));
/*
if (procnum == 1) {
  for (i=1;i<=nborderrow;i++) {
    printf("%d \t%lf %d %d %d \n",i,borderrow[i].Value,borderrow[i].i,
            borderrow[i].j,borderrow[i].NextLBxk);
    fflush(stdout); }}
*/
/* (II) receive a portion of the dense matrix */
printf("(II) receive a portion of the dense matrix \n");
printf("\tln=%d - bksthisproc=%d - nb=%d\n",ln,bksthisproc,nb);
fflush(stdout);
  la = (double *) calloc(((ln*bksthisproc*nb)+1),sizeof(double));
  if (la == NULL)
   {
    printf("can't allocate memory for la[] \n");
    Node_Exit(1);
   } /* end if */
   for (j=0; j<bksthisproc; j++)
    {
     bufsize = ((ln*nb)*sizeof(double));
     CMMD_receive(CMMD_host_node(),RECEIVE_LA,&(la[j*nb*ln]),bufsize);
/*
if (procnum == 1) {
  for (i=0;i<nb;i++) 
   for (k=0;k<ln;k++) {
    l = (j*nb*ln) + ((i*ln)+k);
    printf("%d \t%d %d %lf \n",l,k,i,la[l]);
    fflush(stdout); } }
*/
/*
if (procnum == 1) {
  for (i=(j*nb*ln);i<((j+1)*nb*ln);i++) {
    printf("%d \t%lf \n",i,la[i]);
    fflush(stdout); } }
*/
   } /* end for j */

/* (10) receive the b vector */
  b = (double *) calloc((n+1),sizeof(double));
  if (b == NULL)
   {
    printf("can't allocate memory for b[] \n");
    Node_Exit(1);
   } /* end if */
  CMMD_receive(CMMD_host_node(),RECEIVE_PB,b,((n+1)*sizeof(double)));

/* (III) receive a portion of the lb vector */
printf("(III) receive a portion of the lb vector \n");
fflush(stdout);
 lb = (double *) calloc((ln+1),sizeof(double));
  if (lb == NULL)
   {
    printf("can't allocate memory for lb[] \n");
    Node_Exit(1);
   } /* end if */

  for (j=0; j<bksthisproc; j++)
   {
    bufsize = (nb*sizeof(double));
    CMMD_receive(CMMD_host_node(),RECEIVE_LB,&(lb[((j*nb*nprocs)+(nb*procnum))]),bufsize);
    } /* end for j */

/* (11) receive the row/col locations of the first and last locations in the
   last block */
  CMMD_receive(CMMD_host_node(),RECEIVE_LASTBKST,&lastblockstart,sizeof(int));
  CMMD_receive(CMMD_host_node(),RECEIVE_LASTBKND,&lastblockend,sizeof(int));
printf("lastblockend=%d - lastblockstart=%d \n",lastblockend,lastblockstart);
fflush(stdout);
  nlastblock = lastblockend - lastblockstart + 1;

/* (16) allocate space for the remaining vectors - sparse matrix software */
printf("allocate space for the remaining vectors - sparse matrix software \n");
fflush(stdout);
  k = ((ln*maxbksperproc*nprocs*nb)+nprocs+1);
printf("allocate memory for LBixLBkxBuftmp --- k=%d \n",k);
fflush(stdout);

  LBixLBkxBuftmp = (LBixLBkxBuffer *) calloc(k,sizeof(LBixLBkxBuffer));
  if (LBixLBkxBuftmp == NULL)
   {
    printf("can't allocate memory for LBixLBkxBuftmp \n");
    Node_Exit(1);
   } /* end if */

  LBixLBkxBuf = (LBixLBkxBuffer **) calloc(nprocs,sizeof(LBixLBkxBuffer *));
  if (LBixLBkxBuf== NULL)
   {
    printf("can't allocate memory for LBixLBkxBuf[] \n");
    Node_Exit(1);
   } /* end if */
  for (i=0; i<nprocs; i++)
     LBixLBkxBuf[i] = LBixLBkxBuftmp + i*((ln*maxbksperproc*nb)+1);

  k = ((maxbksperproc*nb*nprocs)+nprocs+1);
  FrbBuftmp = (FRBBuffer *) calloc(k,sizeof(FRBBuffer));
  if (FrbBuftmp == NULL)
   {
    printf("can't allocate memory for FrbBuftmp \n");
    Node_Exit(1);
   } /* end if */

  FrbBuf = (FRBBuffer **) calloc(nprocs,sizeof(FRBBuffer *));
  if (FrbBuf == NULL)
   {
    printf("can't allocate memory for FrbBuf[] \n");
    Node_Exit(1);
   } /* end if */
  for (i=0; i<nprocs; i++)
     FrbBuf[i] = FrbBuftmp + i*((maxbksperproc*nb)+1);

  y = (double *) calloc((n+1),sizeof(double));
  if (y == NULL)
   {
    printf("can't allocate memory for y[] \n");
    Node_Exit(1);
   } /* end if */

  x = (double *) calloc((n+1),sizeof(double));
  if (x == NULL)
   {
    printf("can't allocate memory for x[] \n");
    Node_Exit(1);
   } /* end if */

/* (IV) allocate memory for the remaining variables  - last block */
printf("(IV) allocate memory for the remaining variables \n");
fflush(stdout);
  lx = (double *) calloc((ln+1),sizeof(double));
  if (lx == NULL)
   {
    printf("can't allocate memory for lx[] \n");
    Node_Exit(1);
   } /* end if */
  ly = (double *) calloc((ln+1),sizeof(double));
  if (ly == NULL)
   {
    printf("can't allocate memory for ly[] \n");
    Node_Exit(1);
   } /* end if */
  maxbufsize = (ln*nb*sizeof(double)) + (10*sizeof(int));
  buf2 = (char *) calloc(maxbufsize,sizeof(char));
  if (buf2 == NULL)
   {
    printf("can't allocate memory for buf2[] \n");
    Node_Exit(1);
   } /* end if */
 } /* end if this is an active processor */
} /* end InputData() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
void Factor()
{
#include "AIX.h"

  CMMD_sync_with_host();
  if (procnum < nprocs) 
    {
printf("FactorIndependentBlocks \n");
fflush(stdout);
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     FactorIndependentBlocks();

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_FIB,&node_tim,sizeof(node_timer));
   } /* end if an active processor */

  CMMD_sync_with_host();
  if (procnum < nprocs) 
    {
printf("UpdateLastBlock() \n");
fflush(stdout);
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     UpdateLastBlock();

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_ULB,&node_tim,sizeof(node_timer));
   } /* end if an active processor */

  CMMD_sync_with_host();
  if (procnum < nprocs) 
    {
printf("FactorLastBlock \n");
fflush(stdout);
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     FactorLastBlock();

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_FLB,&node_tim,sizeof(node_timer));
   } /* end if an active processor */

} /* end Factor() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/ 
void FactorIndependentBlocks()
{
  int i,k,l;
  double Lkk;

/* loop over the (nblocks) independent diagonal blocks 
   the borders will be factorized after each diagonal block 
   division by the value of Lkk occurs after a value is updated */

     for (l=1; l<=nblocks; l++)
      {
/* loop along the diagonal in a block */
    for (i=(blocks[l-1].DiagPtr+1); i<=blocks[l].DiagPtr; i++)
        {
         k = i - blocks[0].DiagPtr;
/*
printf("FactorIndependentBlocks() --- l=%d - k=%d \n",l,k);
fflush(stdout);
*/
/* update the value on the diagonal */
/*
printf("\tUpdateLkk(%d)\n",k);
fflush(stdout);
*/
         Lkk = UpdateLkk(k);

/* update the column Lxk */

  /* for nonzero elements in the kth column of the
     strictly lower diagonal porition of this block */
/*
printf("\tUpdateLxk(%d,%lf) \n",k,Lkk);
fflush(stdout);
*/
         UpdateLxk(k,Lkk);

  /* for nonzero elements in the kth column of the
     lower border of this block */
/*
printf("\tUpdateLBxk(%d,%d,%lf) \n",k,l,Lkk);
fflush(stdout);
*/
         UpdateLBxk(k,l,Lkk);

        } /* end for i - loop along diagonal in a block */
      } /* end for l - loop over independent blocks */
} /* end FactorIndependentBlocks() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
double UpdateLkk(k)
  int k;
{
  int locLkj,endLxj;

/* update the value on the diagonal (i=k) */
/* initialize the pointers into the data structures */
  locLkj = (diag[ (k-1) ].PtrLix + 1);
  endLxj = diag[k].PtrLix;

  while(locLkj <= endLxj)
   {
    diag[k].Value -= blockrow[locLkj].Value * blockrow[locLkj].Value;
    locLkj++;
   } /* end while locLkj <= endLxj */

  if (diag[k].Value <= 0)
    {
     printf("OOPS --- (diag[%d].Value <= 0)\n",k);
     fflush(stdout);
     Node_Exit(1);
    } /* end if */
  else
    {
     diag[k].Value = sqrt(diag[k].Value);
     return diag[k].Value;
    } /* end else */
} /* end UpdateLkk() */

 
/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void UpdateLxk(k,Lkk)
  int k;
  double Lkk;
{
  int i;
  int locLik,locLij,locLkj;
  int endLix,endLkx;

  locLik = diag[k].NextLxk;
  while(locLik != 0)
   {
    i = blockrow[locLik].i - blocks[0].DiagPtr;
/*
if(procnum == (nprocs-1))
  {
   printf("\t\tblockrow[%d].i=%d - blockrow[%d].j=%d - i=%d \n",locLik,blockrow[locLik].i,locLik,blockrow[locLik].j,i);
   fflush(stdout); } 
*/
/* initialize the pointers into the data structures */
    locLij = (diag[ (i-1) ].PtrLix + 1);
    locLkj = (diag[ (k-1) ].PtrLix + 1);
    endLix = diag[i].PtrLix;
    endLkx = diag[k].PtrLix;
/* find the first pair of nonzero Lij and Lkj */
/*
if(procnum == (nprocs-1))
  {
   printf("\t\ti=%d - k=%d - locLij=%d locLkj=%d - endLix=%d - endLkx=%d - \n",
         i,k,locLij,locLkj,endLix,endLkx);
   fflush(stdout); } 
*/
    locLixLjx(k,&locLij,&locLkj,endLix,endLkx);
/*
if(procnum == (nprocs-1))
  {
   printf("\t\ti=%d - k=%d - locLij=%d locLkj=%d - endLix=%d - endLkx=%d - \n",
         i,k,locLij,locLkj,endLix,endLkx);
   fflush(stdout); } 
*/
     while( (locLij != 0) && (locLkj != 0) )
     {
/*
if(procnum == (nprocs-1))
  {
   printf("\t\t\tblockrow[%d].i=%d - blockrow[%d].j=%d \n",
          locLik,blockrow[locLik].i,locLik,blockrow[locLik].j);
   printf("\t\t\t\tblockrow[%d].i=%d - blockrow[%d].j=%d \n",
          locLij,blockrow[locLij].i,locLij,blockrow[locLij].j);
   printf("\t\t\t\tblockrow[%d].i=%d - blockrow[%d].j=%d \n",
          locLkj,blockrow[locLkj].i,locLkj,blockrow[locLkj].j);
   fflush(stdout); }
*/
      blockrow[locLik].Value -= blockrow[locLij++].Value * blockrow[locLkj++].Value;
/* find the next pair of nonzero Lij and Lkj */
      locLixLjx(k,&locLij,&locLkj,endLix,endLkx);
     } /* end while (locLij and locLkj != 0) */

    blockrow[locLik].Value /= Lkk;
    locLik = blockrow[locLik].NextLxk;
   } /* end while (locLik !=0) */
} /* end UpdateLxk() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void locLixLjx(k,locLij,locLkj,endLix,endLkx)
  int k,*locLij,*locLkj,endLix,endLkx;
{
  int Search;
  int lkj,lij;

  if ( (*locLij <= endLix) && (*locLkj <= endLkx))
   {
    lkj = blockrow[*locLkj].j - blocks[0].DiagPtr;
    lij = blockrow[*locLij].j - blocks[0].DiagPtr;
    Search = YES;
    while (Search == YES)
     {
      if (lkj == lij)
       {
/*
if(procnum == (nprocs-1))
  {
   printf("\t\t(lkj == lij) --- lkj=%d - lij=%d - k=%d \n",lkj,lij,k);
   fflush(stdout); }
*/
        Search = NO;
       } /* end if lkj == lij */
      else if (lkj < lij)
       {
/*
if(procnum == (nprocs-1))
  {
   printf("\t\t(lkj < lij) --- lkj=%d - lij=%d - k=%d \n",lkj,lij,k);
   fflush(stdout); }
*/
        while ( (lkj < lij) && (lkj < k) && (*locLkj <= endLkx) )
         lkj = blockrow[ ++(*locLkj) ].j - blocks[0].DiagPtr;
/*
if(procnum == (nprocs-1))
  {
   printf("\t\t\t\t(lkj < lij) --- lkj=%d - lij=%d - k=%d \n",lkj,lij,k);
   fflush(stdout);}
*/
        if ( (lkj >= k ) || (*locLkj > endLkx) )
         {
          Search = NO;
          *locLij = *locLkj = 0;
         } /* end if */
       } /* end if lkj < lij */
      else if (lkj > lij)
       {
/*
if(procnum == (nprocs-1))
  {
   printf("\t\t(lkj > lij) --- lkj=%d - lij=%d - k=%d \n",lkj,lij,k);
   fflush(stdout); }
*/
        while ( (lkj > lij) && (lij < k) && (*locLij <= endLix) )
         lij = blockrow[ ++(*locLij) ].j - blocks[0].DiagPtr;
/*
if(procnum == (nprocs-1))
  {
   printf("\t\t\t\t(lkj > lij) --- lkj=%d - lij=%d - k=%d \n",lkj,lij,k);
   fflush(stdout); }
*/
        if ( (lij >= k) || (*locLij > endLix) )
         {
          Search = NO;
          *locLij = *locLkj = 0;
         } /* end if */
       } /* end if lkj > lij */
     } /* end while search == YES */
   } /* end if (*locLij <= endLix) && (*locLkj <= endLkx) */
  else 
   {
    *locLij = *locLkj = 0;
   } /* end else */

} /* end locLixLjx() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void UpdateLBxk(k,l,Lkk)
  int k;
  int l;
  double Lkk;
{
  int i;
  int locLBik,locLBij,locLkj;
  int endLBix,endLkx;
  int LBPtr,EndLBPtr,LBix;

/* initialize the pointers into the data structures */
  LBPtr = (blocks[l-1].LBPtr + 1);
  EndLBPtr = blocks[l].LBPtr;
  LBix = lborder[LBPtr].i;

  locLBik = diag[k].NextLBxk;
  while(locLBik != 0)
   {
    i = borderrow[locLBik].i;
/* find the location of the ith row in the lower border */
    while( (LBix < i) && (LBPtr < EndLBPtr) ) LBix = lborder[++LBPtr].i;
    if (LBix != i)
     {
      printf("LBix != i in UpdateLBxk(%d,%d) LBix=%d i=%d\n",k,l,LBix,i);
      Node_Exit(1);
     } /* end if LBix != i */

    locLBij = (lborder[ (LBPtr - 1) ].PtrLBix + 1);
    locLkj = (diag[ (k-1) ].PtrLix + 1);
    endLBix = lborder[LBPtr].PtrLBix;
    endLkx = diag[k].PtrLix;
/* find the first pair of nonzero LBij and Lkj */
    locLBixLjx(k,&locLBij,&locLkj,endLBix,endLkx);
    while( (locLBij != 0) && (locLkj != 0) )
     {
      borderrow[locLBik].Value -= borderrow[locLBij++].Value * blockrow[locLkj++].Value;
/* find the next pair of nonzero LBij and Lkj */
      locLBixLjx(k,&locLBij,&locLkj,endLBix,endLkx);
     } /* end while (locLBij and locLkj != 0) */

    borderrow[locLBik].Value /= Lkk;
    locLBik = borderrow[locLBik].NextLBxk;
   } /* end while (locLBik !=0) */

} /* end UpdateLBik() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void locLBixLjx(k,locLBij,locLkj,endLBix,endLkx)
  int k,*locLBij,*locLkj,endLBix,endLkx;
{
  int Search;
  int lkj,lbij;

  if ( (*locLBij <= endLBix) && (*locLkj <= endLkx) )
   {
    lkj  = blockrow[*locLkj].j - blocks[0].DiagPtr;
    lbij = borderrow[*locLBij].j - blocks[0].DiagPtr;
    Search = YES;
    while (Search == YES)
     {
      if (lkj == lbij)
       {
        Search = NO;
       } /* end if lkj == lbij */
      else if (lkj < lbij)
       {
        while ( (lkj < lbij) && (lkj < k) && (*locLkj <= endLkx) )
         lkj = blockrow[ ++(*locLkj) ].i - blocks[0].DiagPtr;
        if ( (lkj >= k) || (*locLkj > endLkx) )
         {
          Search = NO;
          *locLBij = *locLkj = 0;
         } /* end if */
       } /* end if lkj < lbij */
      else if (lkj > lbij)
       {
        while ( (lkj > lbij) && (lbij < k) && (*locLBij <= endLBix) )
         lbij = borderrow[ ++(*locLBij) ].j - blocks[0].DiagPtr;
        if ( (lbij >= k) || (*locLBij > endLBix) )
         {
          Search = NO;
          *locLBij = *locLkj = 0;
         } /* end if */
       } /* end if lkj > lbij */
     } /* end while search == YES */
   } /* end if (*locLBij <= endLBix) && (*locLkj <= endLkx) */
  else
   {
    *locLBij = *locLkj = 0;
   } /* end else */

} /* end locLBixLjx() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void UpdateLastBlock()
{
  int i,j,k,l,p,j1,j2,j3,slbcnt;
  int tmpbufsize;
  int iproc,jproc;
  int *m;
  double **SumLBixLBkx,*SumLBixLBkxtemp;
  LBixLBkxBuffer *tmp_rec;
#include "AIX.h"

  m = (int *) calloc((nprocs+1),sizeof(int));
  if (m == NULL)
   {
    printf("can't allocate memory for m[] \n");
    Node_Exit(1);
   } /* end if */
   for (iproc=0;iproc<nprocs;iproc++) m[iproc] = 0;
/*
printf("\tAllocate memory for SumLBixLBkx[][] \n");
fflush(stdout); 
*/
/* allocate memory for SumLBixLBkx[][] */
  SumLBixLBkxtemp = (double *) calloc(((ln+1)*(ln+1)),sizeof(double));
  if (SumLBixLBkxtemp == NULL)
   {
    printf("can't allocate memory for SumLBixLBkxtemp \n");
    Node_Exit(1);
   } /* end if */

  SumLBixLBkx = (double **) calloc((ln+1),sizeof(double *));
  if (SumLBixLBkx == NULL)
   {
    printf("can't allocate memory for SumLBixLBkx[] \n");
    Node_Exit(1);
   } /* end if */
  for (i=0; i<(ln+1); i++)
     SumLBixLBkx[i] = SumLBixLBkxtemp + i*(ln+1);

  for (i=1; i<=ln; i++)
   for (j=1; j<=ln; j++)
    {
     SumLBixLBkx[i][j] = 0.0;
    } /* end for i and j */
/*
printf("\tCalculate and sum the updates to the last block \n");
fflush(stdout); 
*/
/* calculate and sum the updates to the last block */

  slbcnt = 0;
  for (l=1; l<=nblocks; l++)
   {
/*
printf("\t\tl=%d - nblocks=%d \n",l,nblocks);
fflush(stdout); 
*/
    LBixLBkx(l,&slbcnt,SumLBixLBkx);
   } /* end for l */

/* load LBixLBkxBuf */
  for (j=1; j<=ln; j++)
   for (i=j; i<=ln; i++)
    {
     if ( (SumLBixLBkx[i][j] > EPSILON) || (SumLBixLBkx[i][j] < (-1.0 * EPSILON)) )
      {
       iproc = (((j-1)%(nprocs*nb))/nb);
       LBixLBkxBuf[iproc][ m[iproc] ].Value = SumLBixLBkx[i][j];
       LBixLBkxBuf[iproc][ m[iproc] ].i     = (i - 1);
       LBixLBkxBuf[iproc][ m[iproc] ].j     = (j - 1);
/*
if ( (procnum == (nprocs-1)) || (procnum == 0) ) {
   printf("\t\tiproc=%d - m[%d]=%d - i=%d - j=%d - SumLBixLBkx[%d][%d]=%lf \n",
            iproc,iproc,m[iproc],i,j,i,j,SumLBixLBkx[i][j]);
   fflush(stdout); }
*/
       m[iproc]++;
      } /* end if */
    } /* end for i and j */

/* the update buffer only has nonzero values - so mark the
   last value with zeros in the row and column markers */
  for (iproc=0;iproc<nprocs;iproc++)
   {
    LBixLBkxBuf[iproc][ m[iproc] ].Value = 0.0;
    LBixLBkxBuf[iproc][ m[iproc] ].i     = 0;
    LBixLBkxBuf[iproc][ m[iproc] ].j     = 0;
/*
printf("\t\tiproc=%d - m[%d]=%d slbcnt=%d \n",iproc,iproc,m[iproc],slbcnt);
fflush(stdout); 
*/
   } /* end for iproc */

/* update the values in the last block  - this is the portion
   of the Shur complement that updates the last block using
   the values in the borders */


/* send the appropriate data to the other processors - each processor 
   will send data to all other processors when it is it's turn - 
   otherwise the processor will receive.  After receiving data, 
   the processor updates it's values of la( , ).  After sending data, 
   a processor updates it's own values  */

  if (nprocs > 1)
    {
     tmpbufsize = ((ln*maxbksperproc*nb)+1);
     tmp_rec = (LBixLBkxBuffer *)calloc(tmpbufsize,sizeof(LBixLBkxBuffer));
     if (tmp_rec == NULL)
      {
       printf("can't allocate memory for tmp_rec \n");
       Node_Exit(1);
      } /* end if */
     tmpbufsize *= sizeof(LBixLBkxBuffer);
    } /* end if */

  for (iproc=0;iproc<nprocs;iproc++)
   {
    if (iproc == procnum) /* send LBixLBkxBuf[jproc] and update own la( , ) */
     {
      for (jproc=0;jproc<nprocs;jproc++)
       {
        if (jproc != procnum)
         {
/*
printf("\t\t\tiproc=%d - jproc=%d - tmpbufsize=%d - m[%d]=%d - (bytes)=%d \n",
          iproc,jproc,tmpbufsize,jproc,m[jproc],(m[jproc]*sizeof(LBixLBkxBuffer)));
fflush(stdout);
*/
           CMMD_send(jproc,SEND_LBixLBkxBuf,LBixLBkxBuf[jproc],((m[jproc]+1)*sizeof(LBixLBkxBuffer)));
	 } /* end if */
       } /* end for jproc */

/* update la( , ) stored on this processor */
      for (k=0;k<m[iproc];k++)
       {
/*
if(procnum == (nprocs-1))
  {
   printf("\t\t\tiproc=%d - m[%d]=%d - k=%d  \n",
            iproc,iproc,m[iproc],k);
   fflush(stdout); }
*/
        i = LBixLBkxBuf[iproc][k].i;
        j = LBixLBkxBuf[iproc][k].j;
        j1 = (j/(nb*nprocs));
        j2 = (j%nb);
        j3 = (j1*nb)+j2;
/*
if(procnum == (nprocs-1))
  {
   printf("\t\t\ti=%d - j=%d - l=%d - j1=%d - j2=%d - j3=%d  \n",
            i,j,l,j1,j2,j3);
   fflush(stdout); }
*/
        la[ ((j3*ln)+i) ] -= LBixLBkxBuf[procnum][k].Value;
       } /* end for k */
     } /* end for iproc == procnum (send data) */
    else /* receive LBixLBkxBuf[jproc] and update own la( , ) */
     {
      CMMD_receive(iproc,RECEIVE_LBixLBkxBuf,tmp_rec,tmpbufsize);

      k = 0;
      while  ( (tmp_rec[k].i != 0) && (tmp_rec[k].j != 0) && (tmp_rec[k].Value != 0.0) )
       {
        i = tmp_rec[k].i;
        j = tmp_rec[k].j;
        j1 = (j/(nb*nprocs));
        j2 = (j%nb);
        j3 = (j1*nb)+j2;
        la[ ((j3*ln)+i) ] -= tmp_rec[k].Value;
        k++;
       } /* end while */
     } /* end else (receive data) */

   } /* end for iproc */
   free(SumLBixLBkx);
   free(SumLBixLBkxtemp);
   free(tmp_rec);
} /* end UpdateLastBlock() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void LBixLBkx(l,slbcnt,SumLBixLBkx)
  int l,*slbcnt;
  double **SumLBixLBkx;
{

  int locLBij,locLBkj;
  int endLBix,endLBkx;
  int LBPtr1,LBkx,LBPtr2,LBix,EndLBPtr;
/*
printf("\t\t\t\tin LBixLBkx() --- l=%d  \n",l);
fflush(stdout); 
*/
/* initialize the pointers into the data structures */
  LBPtr1 = (blocks[l-1].LBPtr + 1);
  EndLBPtr = blocks[l].LBPtr;

/* loop over all non-zero lower border rows in this block --- 
   visit them in a pair of nested while loops */
/* outer loop */
  while (LBPtr1 <= EndLBPtr)
    {
/*
printf("\t\t\t\t\tLBPtr1=%d - EndLBPtr=%d \n",LBPtr1,EndLBPtr);
fflush(stdout); 
*/
     LBkx = lborder[LBPtr1].i - lastblockstart + 1;
     LBPtr2 = LBPtr1;
/* inner loop */
     while (LBPtr2 <= EndLBPtr)
       {
/*
printf("\t\t\t\t\t\tLBPtr2=%d - EndLBPtr=%d \n",LBPtr2,EndLBPtr);
fflush(stdout); 
*/
        (*slbcnt)++;
        locLBkj = (lborder[ (LBPtr1 - 1) ].PtrLBix + 1);
        endLBkx = lborder[LBPtr1].PtrLBix;
        LBix = lborder[LBPtr2].i - lastblockstart + 1;
        locLBij = (lborder[ (LBPtr2 - 1) ].PtrLBix + 1);
        endLBix = lborder[LBPtr2].PtrLBix;
/*
printf("\t\tLBkx=%d - LBix=%d - locLBkj=%d - endLBkx=%d - locLBij=%d - endLBix=%d\n",
        LBkx,LBix,locLBkj,endLBkx,locLBij,endLBix);
fflush(stdout); 
*/
/* data must exist in this partial row and partial column */
/* find the first pair of nonzero LBij and LBkj */
        locLBixLBkx(&locLBij,&locLBkj,endLBix,endLBkx);
/*
printf("\t\tLBkx=%d - LBix=%d - locLBkj=%d - endLBkx=%d - locLBij=%d - endLBix=%d\n",
        LBkx,LBix,locLBkj,endLBkx,locLBij,endLBix);
fflush(stdout);
*/
        while( (locLBij != 0) && (locLBkj != 0) )
         {
          SumLBixLBkx[LBix][LBkx] += borderrow[locLBij++].Value * borderrow[locLBkj++].Value;
/* find the next pair of nonzero LBij and LBkj */
          locLBixLBkx(&locLBij,&locLBkj,endLBix,endLBkx);
         } /* end while (locLBij and locLBkj != 0) */
        LBPtr2++;
       } /* end inner loop */
     LBPtr1++;
    } /* end outer loop */
  
} /* end LBixLBkx() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void locLBixLBkx(locLBij,locLBkj,endLBix,endLBkx)
  int *locLBij,*locLBkj,endLBix,endLBkx;
{
  int Search;
  int lbkj,lbij;

  if ( (*locLBij <= endLBix) && (*locLBkj <= endLBkx) )
   {
    lbkj = borderrow[*locLBkj].j;
    lbij = borderrow[*locLBij].j;
    Search = YES;
    while (Search == YES)
     {
      if (lbkj == lbij)
       {
/*
if(procnum == (nprocs-1))
  {
   printf("\t\t(lbkj == lbij) --- lbkj=%d - lbij=%d \n",lbkj,lbij);
   fflush(stdout); }
*/
        Search = NO;
       } /* end if lbkj == lbij */
      else if (lbkj < lbij)
       {
/*
if(procnum == (nprocs-1))
  {
   printf("\t\t(lbkj < lbij) --- lbkj=%d - lbij=%d  \n",lbkj,lbij);
   fflush(stdout); }
*/
        while ( (lbkj < lbij) && (*locLBkj <= endLBkx) )
         lbkj = borderrow[ ++(*locLBkj) ].j;
/*
if(procnum == (nprocs-1))
  {
   printf("\t\t\t\t(lbkj < lbij) --- lbkj=%d - lbij=%d \n",lbkj,lbij);
   fflush(stdout);}
*/
        if (*locLBkj > endLBkx)
         {
          Search = NO;
          *locLBij = *locLBkj = 0;
         } /* end if */
       } /* end if lbkj < lbij */
      else if (lbkj > lbij)
       {
/*
if(procnum == (nprocs-1))
  {
   printf("\t\t(lbkj > lbij) --- lbkj=%d - lbij=%d \n",lbkj,lbij);
   fflush(stdout);}
*/
        while ( (lbkj > lbij) && (*locLBij <= endLBix) )
         lbij = borderrow[ ++(*locLBij) ].j;
/*
if(procnum == (nprocs-1))
  {
   printf("\t\t\t\t(lbkj > lbij) --- lbkj=%d - lbij=%d \n",lbkj,lbij);
   fflush(stdout);}
*/
        if (*locLBij > endLBix)
         {
          Search = NO;
          *locLBij = *locLBkj = 0;
         } /* end if */
       } /* end if lbkj > lbij */
     } /* end while search == YES */
   } /* end if (*locLBij <= endLBix) && (*locLBkj <= endLBkx) */
  else
   {
    *locLBij = *locLBkj = 0;
   } /* end else */

} /* end locLBixLBkx() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void FactorLastBlock()
{
  double *work;

/* hgetrf - variables 
   hlltbk(&n,&lda,&nb,la,work,&info,&me,&cols);

   n    is the number of columns in the matrix
   lda  is the first dimension of A (rows)
   nb   is the number of columns in a block
   la   is the square, dense matrix
   work is working buffer
   info is a variable to return the return state of LAPACK routines
   me   is the number of this processor 
   nprocs is the number of processors being used
   cols is the number of columns this processor*/

  work = (double *) calloc(((ln*ln)+1),sizeof(double));
  if (work == NULL)
   {
    printf("can't allocate memory for work[] \n");
    Node_Exit(1);
   } /* end if */
/*
printf("ln=%d - lda=%d - nb=%d - info=%d - procnum=%d - colsthisproc=%d \n",
        ln,ln,nb,info,procnum,colsthisproc);
fflush(stdout);
*/
  hlltbk(&ln,&ln,&nb,la,work,&info,&procnum,&nprocs,&colsthisproc);
  free(work);

} /* end FactorLastBlock() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void ForwardReduction()
{
  int l;
#include"AIX.h"
/* solve for Ly=b */
/* reduce the independent blocks */
  CMMD_sync_with_host();
  if (procnum < nprocs) 
    {
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     for (l=1;l<=nblocks;l++)
      {

       ForwardReduceBlocks(l);

      } /* end for l */

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_FRIB,&node_tim,sizeof(node_timer));
    } /* end if an active processor */

/* reduce this processors portion of the lower border */
  CMMD_sync_with_host();
  if (procnum < nprocs) 
    {
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     ForwardReduceBorder();

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_FRLB,&node_tim,sizeof(node_timer));
    } /* end if an active processor */

/* reduce this processors portion of the last block */
  CMMD_sync_with_host();
  if (procnum < nprocs) 
    {
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     ForwardReduceLastBlock();

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_FR,&node_tim,sizeof(node_timer));
   } /* end if an active processor */

} /* end ForwardReduction() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void ForwardReduceBlocks(l)
  int l;
{
  int i,j;
  int locLij,endLix;

  for (i=(blocks[l-1].DiagPtr+1); i<=blocks[l].DiagPtr; i++)
   {
     j = i - blocks[0].DiagPtr;
     locLij = (diag[j-1].PtrLix + 1);
     endLix = diag[j].PtrLix;
     while (locLij <= endLix)
      {
       b[j] -= blockrow[locLij].Value * y[ (blockrow[locLij].j - blocks[0].DiagPtr) ];
       locLij++;
      } /* end while */
/* the values of Lii are stored on the diagonal */
    y[j] = b[j] / diag[j].Value;
   } /* end for i */

} /* end ForwardReduceBlocks() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void ForwardReduceBorder()
{
  int i,j,k,l,*m;
  int iproc,jproc,tmpbufsize;
  int locLBij,endLBix;
  int LBPtr,EndLBPtr;
  double *frbtmp;  /* frbtmp[] */
  FRBBuffer *tmp_rec;
#include"AIX.h"

  m = (int *) calloc((nprocs+1),sizeof(int));
  if (m == NULL)
   {
    printf("can't allocate memory for m[] \n");
    Node_Exit(1);
   } /* end if */
   for (iproc=0;iproc<nprocs;iproc++) m[iproc] = 0;

  frbtmp = (double *) calloc((nlastblock+1),sizeof(double));
  if (frbtmp == NULL)
   {
    printf("can't allocate memory for frbtmp[] \n");
    Node_Exit(1);
   } /* end if */
  for (i=0; i<=nlastblock; i++) frbtmp[i] = 0.0;

/* reduce the border - except for the last lower diagonal block */
  for (l=1; l<=nblocks; l++)
     {
/* initialize the pointers into the data structures */
      LBPtr = (blocks[l-1].LBPtr + 1);
      EndLBPtr = blocks[l].LBPtr;

/* some lower border rows may be all zeros - so only use
   rows with nonzero values */
/*
printf("some lower border rows may be all zeros - so only use\n");
fflush(stdout);
*/
      while(LBPtr <= EndLBPtr) 
       {
        i = lborder[LBPtr].i - lastblockstart + 1;
        locLBij = (lborder[(LBPtr-1)].PtrLBix + 1);
        endLBix = lborder[LBPtr].PtrLBix;

        while (locLBij <= endLBix)
         {
          frbtmp[i] += borderrow[locLBij].Value * y[ (borderrow[locLBij].j - blocks[0].DiagPtr) ];
          locLBij++;
         } /* end while (locLBij <= endLix) */
        LBPtr++;
       } /* end while (LBPtr <= EndLBPtr) */
     } /* end for l */
/*
printf("place only nonzero values in the buffer to be sent to the control node\n");
fflush(stdout);
*/
/* place only nonzero values in the buffer to be sent to the control node */
  for (i=1; i<=nlastblock; i++)
   {
    if (frbtmp[i] != 0.0)
     {
      iproc = (((i-1)%(nprocs*nb))/nb);
/*
printf("\t\ti=%d - nprocs=%d - nb=%d - iproc=%d - frbtmp[%d]=%lf - ",i,nprocs,nb,iproc,i,frbtmp[i]);
printf("m[%d]=%d \n",iproc,m[iproc]);
fflush(stdout); 
*/
      FrbBuf[iproc][ m[iproc] ].i     = i;
      FrbBuf[iproc][ m[iproc] ].Value = frbtmp[i];
      m[iproc]++;
     } /* end if */
   } /* end for i */
  for (iproc=0;iproc<nprocs;iproc++)
   {
/*
printf("\t\tSet last values to zero ---  m[%d]=%d \n",iproc,m[iproc]);
fflush(stdout);
*/
    FrbBuf[iproc][ m[iproc] ].i     = 0;
    FrbBuf[iproc][ m[iproc] ].Value = 0.0;
   } /* end for i */

/* send the appropriate data to the other processors - each processor 
   will send data to all other processors when it is it's turn - 
   otherwise the processor will receive.  After receiving data, 
   the processor updates it's values of lb( ).  After sending data, 
   a processor updates it's own values  */

  tmpbufsize = ((maxbksperproc*nb)+1);
  tmp_rec = (FRBBuffer *) calloc(tmpbufsize,sizeof(FRBBuffer));
  if (tmp_rec == NULL)
   {
    printf("can't allocate memory for tmp_rec[] \n");
     Node_Exit(1);
   } /* end if */
  tmpbufsize *= sizeof(FRBBuffer);
/*
printf("send the appropriate data to the other processors \n");
fflush(stdout);
*/
  for (iproc=0;iproc<nprocs;iproc++)
   {
    if (iproc == procnum) /* send FrbBuf[jproc] and update own lb( ) */
     {
      for (jproc=0;jproc<nprocs;jproc++)
       {
        if (jproc != procnum)
         {
/*
printf("CMMD_send --- source=%d - destination=%d - tmpbufsize=%d - FrbBuf[%d][0].i=%d - FrbBuf[%d][0].Value=%lf \n",
               iproc,jproc,tmpbufsize,jproc,FrbBuf[jproc][0].i,jproc,FrbBuf[jproc][0].Value);
fflush(stdout);
*/
           CMMD_send(jproc,(SEND_FRBBuf+jproc),FrbBuf[jproc],((m[jproc]+1)*sizeof(FRBBuffer)));
	 } /* end if */
       } /* end for jproc */
/* update lb( ) stored on this processor */
      for (k=0;k<m[iproc];k++)
       {
/*
printf("k=%d -  i=%d - iproc=%d - FrbBuf[%d][%d].i=%d - m[%d]=%d \n",
           k,i,iproc,iproc,k,FrbBuf[iproc][k].i,iproc,m[iproc]);
fflush(stdout); 
*/
        i = FrbBuf[iproc][k].i;
        lb[i] -= FrbBuf[iproc][k].Value;
       } /* end for k */
     } /* end for iproc == procnum (send data) */
    else /* receive FrbBuf[jproc] and update own lb( ) */
     {
/*
printf("\tCMMD_receive --- source=%d - procnum=%d - tmpbufsize=%d -  \n",iproc,procnum,tmpbufsize);
fflush(stdout);
*/
     CMMD_receive(iproc,(RECEIVE_FRBBuf+procnum),tmp_rec,tmpbufsize);
/*
printf("\tCMMD_receive --- tmp_rec[0].i=%d  - tmp_rec[0].Value=%lf \n",
             tmp_rec[0].i,tmp_rec[0].Value);
fflush(stdout);
*/
      k = 0;
      while  ( (tmp_rec[k].i != 0) && (tmp_rec[k].Value != 0.0) )
       {
        i = tmp_rec[k].i;
/*
printf("k=%d -  i=%d - iproc=%d - tmp_rec[%d].i=%d  \n",k,i,iproc,k,tmp_rec[k].i);
fflush(stdout); 
*/
        lb[i] -= tmp_rec[k].Value;
        k++;
       } /* end while */
     } /* end else (receive data) */

   } /* end for iproc */
   free(tmp_rec);
   free(frbtmp);
} /* end ForwardReduceBorder() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void ForwardReduceLastBlock()
{
  int i,j;
  double *work;
/*
printf("in ForwardReduceLastBlock() - 1 \n");
fflush(stdout);
*/
/* hgetrf - variables 
   frmbk(&n,&lda,&nb,la,work,&info,&me,&cols);

   n    is the number of columns in the matrix
   lda  is the first dimension of A (rows)
   nb   is the number of columns in a block
   la   is the square, dense matrix
   ly   is the y vector associated with the last block
   lb   is the b vector associated with the lst block
   work is working buffer
   info is a variable to return the return state of LAPACK routines
   me   is the number of this processor 
   nprocs is the number of processors being used
   cols is the number of columns this processor */

  work = (double *) calloc((ln+1),sizeof(double));
  if (work == NULL)
   {
    printf("can't allocate memory for work[] \n");
    Node_Exit(1);
   } /* end if */
/*
printf("ln=%d - lda=%d - nb=%d - info=%d - procnum=%d - colsthisproc=%d \n",
        ln,ln,nb,info,procnum,colsthisproc);
fflush(stdout);
*/
/*
if(procnum == (nprocs-1)) {
for(j=0;j<colsthisproc;j++)
  for(i=0;i<ln;i++) {
   printf("procnum=%d  ---  la(%d,%d)=%lf \n",procnum,(i+1),(j+1),la[(j*ln)+i]);
   fflush(stdout); }
*/
/*
for(i=0;i<ln;i++) {
   printf("procnum=%d  ---  lb(%d)=%lf \n",procnum,(i+1),lb[i]);
   fflush(stdout); } }
*/
  frmbk(&ln,&ln,&nb,la,ly,lb,work,&info,&procnum,&nprocs,&colsthisproc);
  free(work);
/*
if(procnum == (nprocs-1)) {
for(i=0;i<ln;i++) {
   printf("procnum=%d  ---  ly(%d)=%lf \n",procnum,(i+1),ly[i]);
   fflush(stdout); } }
*/
/*
printf("in ForwardReduceLastBlock() - 2 \n");
fflush(stdout);
*/
} /* end ForwardReduceLastBlock() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void BackwardSubstitution()
{
  int l;
#include"AIX.h"

/* do the backward substitution for the last block */
  CMMD_sync_with_host();
  if (procnum < nprocs) 
    {
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     BackSubLastBlock();

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_BSLB,&node_tim,sizeof(node_timer));
    } /* end if an active processor */

/* do the backward substitution for the remaining blocks */
  CMMD_sync_with_host();
  if (procnum < nprocs) 
    {
     CMMD_node_timer_clear(0);
     CMMD_node_timer_start(0);

     for (l=1;l<=nblocks;l++)
      {

       BackSubBlocks(l);

      } /* end for l */

     CMMD_node_timer_stop(0);
     node_tim.pn_id   = CMMD_self_address();
     node_tim.elapsed = CMMD_node_timer_elapsed(0);
     node_tim.busy    = CMMD_node_timer_busy(0);
     node_tim.idle    = CMMD_node_timer_idle(0);
     CMMD_send(CMMD_host_node(),SEND_TIME_BSIB,&node_tim,sizeof(node_timer));
    } /* end if an active processor */

} /* BackwardSubstitution() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void BackSubLastBlock()
{
  int i,j,i2,proclb;
  double *work;
/*
printf("in BackSubLastBlock() - 1 \n");
fflush(stdout);
*/
/* hgetrf - variables 
   bsmbk(&n,&lda,&nb,la,work,&info,&me,&cols);

   n    is the number of columns in the matrix
   lda  is the first dimension of A (rows)
   nb   is the number of columns in a block
   la   is the square, dense matrix
   work is working buffer
   info is a variable to return the return state of LAPACK routines
   me   is the number of this processor 
   nprocs is the number of processors being used
   cols is the number of columns this processor */

  work = (double *) calloc((ln+1),sizeof(double));
  if (work == NULL)
   {
    printf("can't allocate memory for work[] \n");
    Node_Exit(1);
   } /* end if */

  proclb = (nbks-1)%nprocs;
/*
printf("ln=%d - lda=%d - nb=%d - info=%d - procnum=%d - colsthisproc=%d - bksthisproc=%d - proclb=%d \n",
        ln,ln,nb,info,procnum,colsthisproc,bksthisproc,proclb);
fflush(stdout);
*/
/*
if(procnum == 0) {
for(j=0;j<colsthisproc;j++)
  for(i=0;i<ln;i++) {
   printf("procnum=%d  ---  la(%d,%d)=%lf \n",procnum,(i+1),(j+1),la[(j*ln)+i]);
   fflush(stdout); }
for(i=0;i<ln;i++) {
   printf("procnum=%d  ---  ly(%d)=%lf \n",procnum,(i+1),ly[i]);
   fflush(stdout); } }
*/
  bsmbk(&ln,&ln,&nb,la,ly,lx,work,&info,&procnum,&nprocs,&proclb,
                  &nbks,&colsthisproc,&bksthisproc);
  free(work);
/*
if(procnum == 0) {
for(i=0;i<ln;i++) {
   printf("procnum=%d  ---  lx(%d)=%lf \n",procnum,(i+1),lx[i]);
   fflush(stdout); } }
*/
/*
for(i=0;i<ln;i++) {
   printf("BackSubLastBlock --- \t\tprocnum=%d - lx[%d]=%lf \n",procnum,(i+1),lx[i]);
   fflush(stdout); }
*/
} /* BackSubLastBlock() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void BackSubBlocks(l)
  int l;
{
  int i,j;
  int LBPtr,EndLBPtr;
  int locLBij,endLBix;
  int locLij,endLix;

/* update y[] using the values in the lth border section */
/*
printf("BackSubBlocks(%d) --- update y[] using the values in the %dth border section \n",l,l);
fflush(stdout);
*/
/* initialize the pointers into the data structures */
  LBPtr = blocks[l].LBPtr;
  EndLBPtr = (blocks[l-1].LBPtr + 1);
/*
printf("BackSubBlocks(%d) --- initialize the pointers into the data structures \n",l);
fflush(stdout);
*/
/* some lower border columns may be all zeros - so only use
   columns with nonzero values */
  while (LBPtr >= EndLBPtr)
   {
    i = lborder[LBPtr].i - lastblockstart;
    locLBij = (lborder[ (LBPtr - 1) ].PtrLBix + 1);
    endLBix = lborder[LBPtr].PtrLBix;

    while (locLBij <= endLBix)
     {
      y[ (borderrow[locLBij].j - blocks[0].DiagPtr)] -= borderrow[locLBij].Value * lx[i];
      locLBij++;
     } /* end while (locLBij <= endLBix) */
    LBPtr--;
   } /* end while (LBPtr >= EndLBPtr) */

/* back substitute for the lth block */

  for (i=blocks[l].DiagPtr; i>=(blocks[l-1].DiagPtr+1); i--)
   {
    j = i - blocks[0].DiagPtr;
/*
printf("BackSubBlocks(%d) --- before divide --- i=%d - j=%d - x[%d]=%lf - y[%d]=%lf - diag[%d]=%lf \n",
         l,i,j,j,x[j],j,y[j],j,diag[j].Value);
fflush(stdout);
*/
/* This is Choleski factorization, and values on the diagonal need not be equal to 1.0 */
    x[j] = y[j] / diag[j].Value;

    locLij = (diag[j-1].PtrLix+1);
    endLix = diag[j].PtrLix;
/*
printf("BackSubBlocks(%d) --- locLij=%d - endLix=%d -\n",l,locLij,endLix);
fflush(stdout);
*/
    while (locLij <= endLix)
     {
/*
printf("BackSubBlocks(%d) --- blockrow[%d].i=%d - blocks[0].DiagPtr=%d - blockrow[%d].Value=%lf - x[%d]=%lf - \n",
                 l,locLij,blockrow[locLij].i,blocks[0].DiagPtr,locLij,blockrow[locLij].Value,j,x[j]);
fflush(stdout);
*/
      y[ (blockrow[locLij].i - blocks[0].DiagPtr) ] -= blockrow[locLij].Value * x[j];
      locLij++;
     } /* end while */
   } /* end for i */
/*
printf("BackSubBlocks(%d) --- FINISHED \n",l);
fflush(stdout);
*/

} /* end BackSubBlocks() */
 

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void bcastdoublesd(ni,nj,am,dir,msgid)
int 
    *ni,                  /* (input ) row dimension of a     */
    *nj,                  /* (input ) column dimension of a  */
    *msgid;               /* (input ) message identifier     */ 
double
    *am;                  /* (input ) double array of data  */
char
    *dir;                 /* direction to traverse conceptional ring */
{
  int i,j,*hbuf,bufsize,destproc,bytessent;
  double *dbuf;
#include"AIX.h"

  hbuf = (int *) buf2;
  dbuf = (double *) (buf2+(5*sizeof(int)));

  bufsize = (((*ni) * (*nj))*sizeof(double)) + (5*sizeof(int));
/*
printf("enter bcastdoublesd --- ni=%d - nj=%d - *dir=%c - bufsize=%d - *msgid=%d \n",
        *ni,*nj,*dir,bufsize,*msgid); 
fflush(stdout);
*/
/* pack header */
  hbuf[0] = *ni;
  hbuf[1] = *nj;
  hbuf[2] = procnum;
  hbuf[3] = bufsize;

/* pack matrix in column major form (as stored) */
  for (j=1; j<=(*nj); j++)
   for (i=1; i<=(*ni); i++)
    dbuf[((((j-1)*(*nj))+i)-1)] = am[((((j-1)*(*nj))+i)-1)];

/* send the buffer to the next processor in a bi-directional conceptional ring */

  if( (*dir == 'N') || (*dir == 'n') )
   destproc = nextinring(&procnum);
  else if( (*dir == 'P') || (*dir == 'p') )
   destproc = previnring(&procnum);

  CMMD_send(destproc,*msgid,hbuf,bufsize);
  bytessent = bufsize;
/*
printf("send in bcastdoublesd --- destproc=%d - bufsize=%d - bytessent=%d - *msgid=%d \n",
        destproc,bufsize,bytessent,*msgid);
fflush(stdout);
*/
/*
if ( (*dir == 'N') || (*dir == 'n') ) {
  for (j=1; j<=(*nj); j++)
   for (i=1; i<=(*ni); i++) {
    printf("\t\tprocnum=%d - am(%d,%d)=%lf \n",procnum,i,j,am[((((j-1)*(*nj))+i)-1)]);
    fflush(stdout); } }
*/
/*
if ( (*dir == 'P') || (*dir == 'p') ) {
  for (j=1; j<=(*nj); j++)
   for (i=1; i<=(*ni); i++) {
    printf("\t\tprocnum=%d - am(%d,%d)=%lf \n",procnum,i,j,am[((((j-1)*(*nj))+i)-1)]);
    fflush(stdout); } }
*/
} /* end bcastdoublesd() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void bcastdoublerc(ni,nj,am,dir,msgid)
int 
    *ni,                  /* (output) row dimension of a     */
    *nj,                  /* (output) column dimension of a  */
    *msgid;               /* (input ) message identifier     */ 
double
    *am;                  /* (output) double array of data  */
char
    *dir;                 /* direction to traverse conceptional ring */
{
  int i,j,*hbuf,bufsize;
  int srcproc,destproc,origproc,bytessent,bytesrec;
  double *dbuf;
#include"AIX.h"

/* receive the buffer from the previous processor in a conceptional bidirectional ring */
/*
printf("enter bcastdoublerc --- srcproc=%d - maxbufsize=%d - *dir=%c - *msgid=%d \n",
        srcproc,maxbufsize,*dir,*msgid);
fflush(stdout);
*/
  hbuf = (int *) buf2;
  dbuf = (double *) (buf2+(5*sizeof(int)));

  if( (*dir == 'N') || (*dir == 'n') )
   srcproc = previnring(&procnum);
  else if( (*dir == 'P') || (*dir == 'p') )
   srcproc = nextinring(&procnum);
  CMMD_receive(srcproc,*msgid,hbuf,maxbufsize);
  bytesrec = maxbufsize; /*was CMMD_bytes_received();*/
/*
printf("receive in bcastdoublerc --- srcproc=%d - bufsize=%d - bytesrec=%d - *msgid=%d \n",
        srcproc,bufsize,bytesrec,*msgid);
fflush(stdout);
*/
/* unpack header */
  *ni = hbuf[0];
  *nj = hbuf[1];
  origproc = hbuf[2];
  bufsize = hbuf[3];
/*
printf("*ni=%d - *nj=%d - srcproc=%d - hbuf[3]=%d \n",*ni,*nj,srcproc,hbuf[3]);
fflush(stdout);
*/
/* if necessary, continue broadcast  - CMMD_bc has not been
   used because it is blocking */

  if( (*dir == 'N') || (*dir == 'n') )
   destproc = nextinring(&procnum);
  else if( (*dir == 'P') || (*dir == 'p') )
   destproc = previnring(&procnum);
  if(origproc != destproc)
    {
     CMMD_send(destproc,*msgid,hbuf,bufsize);
     bytessent = bufsize;
/*
printf("send in bcastdoublerc --- destproc=%d - origproc=%d - bufsize=%d - bytessent=%d - *msgid=%d \n",
        destproc,origproc,bufsize,bytessent,*msgid);
fflush(stdout);
*/
    } /* end if */

/* unpack the buffer into the matrix in column major form */
  for (j=1; j<=(*nj); j++)
   for (i=1; i<=(*ni); i++)
    am[((((j-1)*(*nj))+i)-1)] = dbuf[((((j-1)*(*nj))+i)-1)];

} /* end bcastdoublerc() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
int nextinring(iproc)
  int *iproc;
{
  return ((*iproc+1)%nprocs);

} /* end nextinring() */

/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
int previnring(iproc)
  int *iproc;
{
  return ((*iproc+nprocs-1)%nprocs);

} /* end previnring() */


/*<<<<<<<<<<<<<<<<<<<<---------->>>>>>>>>>>>>>>>>>>>*/
void Node_Exit(i)
  int i;
{
  int inf = 1073741824;
  double e1;

  e1 = la[inf];
  exit(i);

} /* Node_Exit() */





